# -*- coding: utf-8 -*-
#
# Copyright © keithley2600 Project Contributors
# Licensed under the terms of the MIT License
# (see keithley2600/__init__.py for details)


# Warning: PROPERTIES, CONSTANTS, FUNCTIONS and CLASSES must be mutually
# exclusive. A string used for a PROPERTY cannot be used to create a class
# later, when the property already exists (and vice versa).

"""
Submodule listing accepted keithley TSP commands.

"""

WARNING = """Warning: PROPERTIES, CONSATNTS, FUNCTIONS must
not contain elements from and CLASSES. E.g., once a property is created,
it cannot be used as a class anymore"""

CONSTANTS = [
            'CAPACITY',
            'ARMED_EVENT_ID',
            'MEASURE_COMPLETE_EVENT_ID',
            'PULSE_COMPLETE_EVENT_ID',
            'SOURCE_COMPLETE_EVENT_ID',
            'SWEEP_COMPLETE_EVENT_ID',
            'SWEEPING_EVENT_ID',
            'IDLE_EVENT_ID',
            'EVENT_ID',
            'OUTPUT_OFF',
            'OUTPUT_ON',
            'OUTPUT_HIGH_Z',
            'OUTPUT_DCAMPS',
            'OUTPUT_DCVOLTS',
            'MEASURE_DCAMPS',
            'MEASURE_DCVOLTS',
            'MEASURE_OHMS',
            'MEASURE_WATTS',
            'DISABLE',
            'ENABLE',
            'SENSE_LOCAL',
            'SENSE_REMOTE',
            'SENSE_CALA',
            'SOURCE_IDLE',
            'SOURCE_HOLD',
            'AUTORANGE_OFF',
            'AUTORANGE_ON',
            'AUTORANGE_FOLLOW_LIMIT',
            'ON',
            'OFF',
            ]

FUNCTIONS = [
            'beep',
            'bitand',
            'bitor',
            'bitxor',
            'clear',
            'get',
            'getfield',
            'set',
            'setfield',
            'test',
            'toggle',
            'clear',
            'clearcache',
            'add',
            'clear',
            'next',
            'delay',
            'readbit',
            'readport',
            'assert',
            'clear',
            'release',
            'reset',
            'wait',
            'writebit',
            'writeport',
            'clear',
            'getannunciators',
            'getcursor',
            'getlastkey',
            'gettext',
            'inputvalue',
            'add',
            'catalog',
            'delete',
            'menu',
            'prompt',
            'sendkey',
            'setcursor',
            'settext',
            'clear',
            'wait',
            'waitkey',
            'clear',
            'next',
            'all',
            'clear',
            'next',
            'exit',
            'chdir',
            'cwd',
            'is_dir',
            'is_file',
            'mkdir',
            'readdir',
            'rmdir',
            'gettimezone',
            'gm_isweep',
            'gm_vsweep',
            'i_leakage_measure',
            'i_leakage_threshold',
            'InitiatePulseTest',
            'InitiatePulseTestDual',
            'close',
            'flush',
            'input',
            'open',
            # 'output',  # XXX: not supported because of conflict with smuX.source.output
            'read',
            'type',
            'write',
            'applysettings',
            'reset',
            'restoredefaults',
            'assert',
            'clear',
            'connect',
            'disconnect',
            'wait',
            'reset',
            'makegetter',
            'makesetter',
            'meminfo',
            'execute',
            'getglobal',
            'setglobal',
            'opc',
            'remove',
            'rename',
            'print',
            'printbuffer',
            'printnumber',
            'QueryPulseConfig',
            'reset',
            'savebuffer',
            'delete',
            'catalog',
            'load',
            'new',
            'newautorun',
            'restore',
            'run',
            'catalog',
            'list',
            'run',
            'save',
            'read',
            'write',
            'settime',
            'settimezone',
            'recall',
            'save',
            'abort',
            'getstats',
            'recalculatestats',
            'lock',
            'restore',
            'save',
            'unlock',
            'calibratehi',
            'calibratelo',
            'check',
            'makebuffer',
            'calibratev',
            'calibratei',
            'overlappedv',
            'overlappedi',
            'overlappedr',
            'overlappedp',
            'measurevandstep',
            'measureiandstep',
            'measurerandstep',
            'measurepandstep',
            'reset',
            'savebuffer',
            'calibratev',
            'calibratei',
            'set',
            'set',
            'initiate',
            'set',
            'linearv',
            'lineari',
            'listv',
            'listi',
            'logv',
            'logi',
            'set',
            'reset',
            't',
            'reset',
            'clear',
            'reset',
            'wait',
            'clear',
            'clear',
            'reset',
            'wait',
            'wait',
            'readbit',
            'readport',
            'reset',
            'assert',
            'clear',
            'release',
            'reset',
            'wait',
            'writebit',
            'writeport',
            'clear',
            'connect',
            'disconnect',
            'execute',
            'idn',
            'read',
            'readavailable',
            'reset',
            'termination',
            'abort',
            'rbtablecopy',
            'runscript',
            'write',
            'add',
            'catalog',
            'delete',
            'get',
            'waitcomplete',
            'delete',
            'get',
            'waitcomplete',
            'v',
            'i',
            'iv',
            'r',
            'p',
            ]

PROPERTIES = [
            'enable',
            'appendmode',
            'basetimestamp',
            'cachemode',
            'capacity',
            'collectsourcevalues',
            'collecttimestamps',
            'fillcount',
            'fillmode',
            'measurefunctions',
            'measureranges',
            'n',
            'readings',
            'sourcefunctions',
            'sourceoutputstates',
            'sourceranges',
            'sourcevalues',
            'statuses',
            'timestampresolution',
            'timestamps',
            'count',
            'mode',
            'overrun',
            'pulsewidth',
            'stimulus',
            'writeprotect',
            'locallockout',
            'numpad',
            'screen',
            'digits',
            'func',
            'func',
            'overrun',
            'count',
            'count',
            'enable',
            'overwritemethod',
            'asciiprecision',
            'byteorder',
            'data',
            'address',
            'autoconnect',
            'address[N]',
            'domain',
            'dynamic',
            'hostname',
            'verify',
            'duplex',
            'gateway',
            'ipaddress',
            'method',
            'speed',
            'subnetmask',
            'linktimeout',
            'lxidomain',
            'nagle',
            'address[N]',
            'name',
            'duplex',
            'gateway',
            'ipaddress',
            'macaddress',
            'dst',
            'rawsocket',
            'telnet',
            'vxi11',
            'speed',
            'subnetmask',
            'timedwait',
            # 'connected', # XXX: conflicts with connected attribute of Keithley2600Base
            'ipaddress',
            'mode',
            'overrun',
            'protocol',
            'pseudostate',
            'stimulus',
            'autolinefreq',
            'description',
            'linefreq',
            'model',
            'password',
            'passwordmode',
            'prompts',
            'prompts4882',
            'revision',
            'serialno',
            'showerrors',
            'anonymous',
            'autorun',
            'name',
            'baud',
            'databits',
            'flowcontrol',
            'parity',
            'poweron',
            'adjustdate',
            'date',
            'due',
            'password',
            'polarity',
            'state',
            'speed',
            'threshold',
            'analogfilter',
            'autorangev',
            'autorangei',
            'autozero',
            'count',
            'delay',
            'delayfactor',
            'count',
            'enable',
            'type',
            'highcrangedelayfactor',
            'interval',
            'lowrangev',
            'lowrangei',
            'nplc',
            'rangev',
            'rangei',
            'enablev',
            'enablei',
            'enabler',
            'enablep',
            'levelv',
            'leveli',
            'levelr',
            'levelp',
            'sense',
            'autorangev',
            'autorangei',
            'compliance',
            'delay',
            'func',
            'highc',
            'limitv',
            'limiti',
            'lowrangev',
            'lowrangei',
            'offfunc',
            'offlimitv',
            'offlimiti',
            'offmode',
            'output',
            'outputenableaction',
            'settling',
            'sink',
            'count',
            'stimulus',
            'autoclear',
            'count',
            'action',
            'stimulus',
            'action',
            'action',
            'stimulus',
            'action',
            'stimulus',
            'condition',
            'node_enable',
            'node_event',
            'request_enable',
            'request_event',
            'orenable',
            'overrun',
            'stimulus',
            'count',
            'delay',
            'delaylist',
            'overrun',
            'passthrough',
            'stimulus',
            'group',
            'master',
            'state',
            'mode',
            'overrun',
            'pulsewidth',
            'stimulus',
            'writeprotect',
            'timeout',
            'abortonconnect',
            'condition',
            'enable',
            'event',
            'ntr',
            'ptr',
            ]

CLASSES = [
            'arm',
            'beeper',
            'bit',
            'blender',
            'buffer',
            'bufferVar',
            'buffer_available',
            'cal',
            'calibrating',
            'calibration',
            'config',
            'contact',
            'current_limit',
            'dataqueue',
            'digio',
            'display',
            'dns',
            'endpulse',
            'endsweep',
            'errorqueue',
            'eventlog',
            'factory',
            'filter',
            'format',
            'fs',
            'gpib',
            'instrument',
            'io',
            'lan',
            'limit',
            'loadmenu',
            'localnode',
            'measure',
            'measurement',
            'measuring',
            'node',
            'nvbuffer1',
            'nvbuffer2',
            'operation',
            'os',
            'over_temperature',
            'port',
            'questionable',
            'reading_overflow',
            'rel',
            'remote',
            'script',
            'serial',
            'setup',
            'smua',
            'smub',
            'source',
            'standard',
            'status',
            'sweeping',
            'system',
            'system2',
            'system3',
            'system4',
            'system5',
            'timer',
            'trigger',
            'trigger_blender',
            'trigger_overrrun',
            'trigger_overrun',
            'trigger_timer',
            'tsp',
            'tsplink',
            'tspnet',
            'unstable_output',
            'user',
            'userstring',
            'voltage_limit',
            ]

ALL = [
 'InitiatePulseTest',
 'InitiatePulseTestDual',
 'QueryPulseConfig',
 'beeper.beep',
 'beeper.enable',
 'bit.bitand',
 'bit.bitor',
 'bit.bitxor',
 'bit.clear',
 'bit.get',
 'bit.getfield',
 'bit.set',
 'bit.setfield',
 'bit.test',
 'bit.toggle',
 'bufferVar.appendmode',
 'bufferVar.basetimestamp',
 'bufferVar.cachemode',
 'bufferVar.capacity',
 'bufferVar.clear',
 'bufferVar.clearcache',
 'bufferVar.collectsourcevalues',
 'bufferVar.collecttimestamps',
 'bufferVar.fillcount',
 'bufferVar.fillmode',
 'bufferVar.measurefunctions',
 'bufferVar.measureranges',
 'bufferVar.n',
 'bufferVar.readings',
 'bufferVar.sourcefunctions',
 'bufferVar.sourceoutputstates',
 'bufferVar.sourceranges',
 'bufferVar.sourcevalues',
 'bufferVar.statuses',
 'bufferVar.timestampresolution',
 'bufferVar.timestamps',
 'dataqueue.CAPACITY',
 'dataqueue.add',
 'dataqueue.clear',
 'dataqueue.count',
 'dataqueue.next',
 'delay',
 'digio.readbit',
 'digio.readport',
 'digio.trigger.assert',
 'digio.trigger.clear',
 'digio.trigger.release',
 'digio.trigger.reset',
 'digio.trigger.wait',
 'digio.trigger[N].EVENT_ID',
 'digio.trigger[N].mode',
 'digio.trigger[N].overrun',
 'digio.trigger[N].pulsewidth',
 'digio.trigger[N].stimulus',
 'digio.writebit',
 'digio.writeport',
 'digio.writeprotect',
 'display.clear',
 'display.getannunciators',
 'display.getcursor',
 'display.getlastkey',
 'display.gettext',
 'display.inputvalue',
 'display.loadmenu.add',
 'display.loadmenu.catalog',
 'display.loadmenu.delete',
 'display.locallockout',
 'display.menu',
 'display.numpad',
 'display.prompt',
 'display.screen',
 'display.sendkey',
 'display.setcursor',
 'display.settext',
 'display.smua.digits',
 'display.smua.limit.func',
 'display.smua.measure.func',
 'display.smub.digits',
 'display.smub.limit.func',
 'display.smub.measure.func',
 'display.trigger.EVENT_ID',
 'display.trigger.clear',
 'display.trigger.overrun',
 'display.trigger.wait',
 'display.waitkey',
 'errorqueue.clear',
 'errorqueue.count',
 'errorqueue.next',
 'eventlog.all',
 'eventlog.clear',
 'eventlog.count',
 'eventlog.enable',
 'eventlog.next',
 'eventlog.overwritemethod',
 'exit',
 'format.asciiprecision',
 'format.byteorder',
 'format.data',
 'fs.chdir',
 'fs.cwd',
 'fs.is_dir',
 'fs.is_file',
 'fs.mkdir',
 'fs.readdir',
 'fs.rmdir',
 'gettimezone',
 'gm_isweep',
 'gm_vsweep',
 'gpib.address',
 'i_leakage_measure',
 'i_leakage_threshold',
 'io.close',
 'io.flush',
 'io.input',
 'io.open',
 #  'io.output', # XXX: not supported because of conflict with smuX.source.output
 'io.read',
 'io.type',
 'io.write',
 'kei.*',
 'lan.applysettings',
 'lan.autoconnect',
 'lan.config.dns.address[N]',
 'lan.config.dns.domain',
 'lan.config.dns.dynamic',
 'lan.config.dns.hostname',
 'lan.config.dns.verify',
 'lan.config.duplex',
 'lan.config.gateway',
 'lan.config.ipaddress',
 'lan.config.method',
 'lan.config.speed',
 'lan.config.subnetmask',
 'lan.linktimeout',
 'lan.lxidomain',
 'lan.nagle',
 'lan.reset',
 'lan.restoredefaults',
 'lan.status.dns.address[N]',
 'lan.status.dns.name',
 'lan.status.duplex',
 'lan.status.gateway',
 'lan.status.ipaddress',
 'lan.status.macaddress',
 'lan.status.port.dst',
 'lan.status.port.rawsocket',
 'lan.status.port.telnet',
 'lan.status.port.vxi11',
 'lan.status.speed',
 'lan.status.subnetmask',
 'lan.timedwait',
 'lan.trigger.assert',
 'lan.trigger.clear',
 'lan.trigger.connect',
 'lan.trigger.disconnect',
 'lan.trigger.wait',
 'lan.trigger[N].EVENT_ID',
 # 'lan.trigger[N].connected',  # XXX: conflicts with connected attribute of Keithley2600Base
 'lan.trigger[N].ipaddress',
 'lan.trigger[N].mode',
 'lan.trigger[N].overrun',
 'lan.trigger[N].protocol',
 'lan.trigger[N].pseudostate',
 'lan.trigger[N].stimulus',
 'localnode.autolinefreq',
 'localnode.description',
 'localnode.linefreq',
 'localnode.model',
 'localnode.password',
 'localnode.passwordmode',
 'localnode.prompts',
 'localnode.prompts4882',
 'localnode.reset',
 'localnode.revision',
 'localnode.serialno',
 'localnode.showerrors',
 'makegetter',
 'makesetter',
 'meminfo',
 'node.execute',
 'node.getglobal',
 'node.setglobal',
 'opc',
 'os.remove',
 'os.rename',
 'print',
 'printbuffer',
 'printnumber',
 'reset',
 'savebuffer',
 'script.anonymous',
 'script.delete',
 'script.factory.catalog',
 'script.load',
 'script.new',
 'script.newautorun',
 'script.restore',
 'script.run',
 'script.user.catalog',
 'serial.baud',
 'serial.databits',
 'serial.flowcontrol',
 'serial.parity',
 'serial.read',
 'serial.write',
 'settime',
 'settimezone',
 'setup.poweron',
 'setup.recall',
 'setup.save',
 'smua.abort',
 'smua.buffer.getstats',
 'smua.buffer.recalculatestats',
 'smua.cal.adjustdate',
 'smua.cal.date',
 'smua.cal.due',
 'smua.cal.lock',
 'smua.cal.password',
 'smua.cal.polarity',
 'smua.cal.restore',
 'smua.cal.save',
 'smua.cal.state',
 'smua.cal.unlock',
 'smua.contact.calibratehi',
 'smua.contact.calibratelo',
 'smua.contact.check',
 'smua.contact.r',
 'smua.contact.speed',
 'smua.contact.threshold',
 'smua.makebuffer',
 'smua.measure.v',
 'smua.measure.i',
 'smua.measure.analogfilter',
 'smua.measure.autorangev',
 'smua.measure.autorangei',
 'smua.measure.autozero',
 'smua.measure.calibratev',
 'smua.measure.calibratei',
 'smua.measure.count',
 'smua.measure.delay',
 'smua.measure.delayfactor',
 'smua.measure.filter.count',
 'smua.measure.filter.enable',
 'smua.measure.filter.type',
 'smua.measure.highcrangedelayfactor',
 'smua.measure.interval',
 'smua.measure.lowrangev',
 'smua.measure.lowrangei',
 'smua.measure.nplc',
 'smua.measure.overlappedv',
 'smua.measure.overlappedi',
 'smua.measure.overlappedr',
 'smua.measure.overlappedp',
 'smua.measure.rangev',
 'smua.measure.rangei',
 'smua.measure.rel.enablev',
 'smua.measure.rel.enablei',
 'smua.measure.rel.enabler',
 'smua.measure.rel.enablep',
 'smua.measure.rel.levelv',
 'smua.measure.rel.leveli',
 'smua.measure.rel.levelr',
 'smua.measure.rel.levelp',
 'smua.measurevandstep',
 'smua.measureiandstep',
 'smua.measurerandstep',
 'smua.measurepandstep',
 'smua.nvbufferY',
 'smua.nvbufferY.readings[N]',
 'smua.reset',
 'smua.savebuffer',
 'smua.sense',
 'smua.source.autorangei',
 'smua.source.autorangev',
 'smua.source.calibratei',
 'smua.source.calibratev',
 'smua.source.compliance',
 'smua.source.delay',
 'smua.source.func',
 'smua.source.highc',
 'smua.source.leveli',
 'smua.source.levelv',
 'smua.source.limiti',
 'smua.source.limitv',
 'smua.source.lowrangei',
 'smua.source.lowrangev',
 'smua.source.offfunc',
 'smua.source.offlimiti',
 'smua.source.offlimitv',
 'smua.source.offmode',
 'smua.source.output',
 'smua.source.outputenableaction',
 'smua.source.rangev',
 'smua.source.rangei',
 'smua.source.settling',
 'smua.source.sink',
 'smua.trigger.ARMED_EVENT_ID',
 'smua.trigger.IDLE_EVENT_ID',
 'smua.trigger.MEASURE_COMPLETE_EVENT_ID',
 'smua.trigger.PULSE_COMPLETE_EVENT_ID',
 'smua.trigger.SOURCE_COMPLETE_EVENT_ID',
 'smua.trigger.SWEEPING_EVENT_ID',
 'smua.trigger.SWEEP_COMPLETE_EVENT_ID',
 'smua.trigger.arm.count',
 'smua.trigger.arm.set',
 'smua.trigger.arm.stimulus',
 'smua.trigger.autoclear',
 'smua.trigger.count',
 'smua.trigger.endpulse.action',
 'smua.trigger.endpulse.set',
 'smua.trigger.endpulse.stimulus',
 'smua.trigger.endsweep.action',
 'smua.trigger.initiate',
 'smua.trigger.measure.i',
 'smua.trigger.measure.v',
 'smua.trigger.measure.r',
 'smua.trigger.measure.p',
 'smua.trigger.measure.action',
 'smua.trigger.measure.set',
 'smua.trigger.measure.stimulus',
 'smua.trigger.source.action',
 'smua.trigger.source.limiti',
 'smua.trigger.source.limitv',
 'smua.trigger.source.lineari',
 'smua.trigger.source.linearv',
 'smua.trigger.source.listi',
 'smua.trigger.source.listv',
 'smua.trigger.source.logY',
 'smua.trigger.source.logi',
 'smua.trigger.source.logv',
 'smua.trigger.source.set',
 'smua.trigger.source.stimulus',
 'smub.abort',
 'smub.buffer.getstats',
 'smub.buffer.recalculatestats',
 'smub.cal.adjustdate',
 'smub.cal.date',
 'smub.cal.due',
 'smub.cal.lock',
 'smub.cal.password',
 'smub.cal.polarity',
 'smub.cal.restore',
 'smub.cal.save',
 'smub.cal.state',
 'smub.cal.unlock',
 'smub.contact.calibratehi',
 'smub.contact.calibratelo',
 'smub.contact.check',
 'smub.contact.r',
 'smub.contact.speed',
 'smub.contact.threshold',
 'smub.makebuffer',
 'smub.measure.Y',
 'smub.measure.analogfilter',
 'smub.measure.autorangeY',
 'smub.measure.autozero',
 'smub.measure.calibrateY',
 'smub.measure.count',
 'smub.measure.delay',
 'smub.measure.delayfactor',
 'smub.measure.filter.count',
 'smub.measure.filter.enable',
 'smub.measure.filter.type',
 'smub.measure.highcrangedelayfactor',
 'smub.measure.interval',
 'smub.measure.lowrangeY',
 'smub.measure.nplc',
 'smub.measure.overlappedY',
 'smub.measure.rangeY',
 'smub.measure.rel.enableY',
 'smub.measure.rel.levelY',
 'smub.measureYandstep',
 'smub.nvbufferY',
 'smub.nvbufferY.readings[N]',
 'smub.reset',
 'smub.savebuffer',
 'smub.sense',
 'smub.source.autorangeY',
 'smub.source.calibrateY',
 'smub.source.compliance',
 'smub.source.delay',
 'smub.source.func',
 'smub.source.highc',
 'smub.source.levelY',
 'smub.source.limitY',
 'smub.source.lowrangeY',
 'smub.source.offfunc',
 'smub.source.offlimitY',
 'smub.source.offmode',
 'smub.source.output',
 'smub.source.outputenableaction',
 'smub.source.rangeY',
 'smub.source.settling',
 'smub.source.sink',
 'smub.trigger.ARMED_EVENT_ID',
 'smub.trigger.IDLE_EVENT_ID',
 'smub.trigger.MEASURE_COMPLETE_EVENT_ID',
 'smub.trigger.PULSE_COMPLETE_EVENT_ID',
 'smub.trigger.SOURCE_COMPLETE_EVENT_ID',
 'smub.trigger.SWEEPING_EVENT_ID',
 'smub.trigger.SWEEP_COMPLETE_EVENT_ID',
 'smub.trigger.arm.count',
 'smub.trigger.arm.set',
 'smub.trigger.arm.stimulus',
 'smub.trigger.autoclear',
 'smub.trigger.count',
 'smub.trigger.endpulse.action',
 'smub.trigger.endpulse.set',
 'smub.trigger.endpulse.stimulus',
 'smub.trigger.endsweep.action',
 'smub.trigger.initiate',
 'smub.trigger.measure.Y',
 'smub.trigger.measure.action',
 'smub.trigger.measure.set',
 'smub.trigger.measure.stimulus',
 'smub.trigger.source.action',
 'smub.trigger.source.limitY',
 'smub.trigger.source.linearY',
 'smub.trigger.source.listY',
 'smub.trigger.source.logY',
 'smub.trigger.source.set',
 'smub.trigger.source.stimulus',
 'status.condition',
 'status.measurement.*',
 'status.measurement.buffer_available.*',
 'status.measurement.current_limit.*',
 'status.measurement.instrument.*',
 'status.measurement.instrument.smua.*',
 'status.measurement.instrument.smub.*',
 'status.measurement.reading_overflow.*',
 'status.measurement.voltage_limit.*',
 'status.node_enable',
 'status.node_event',
 'status.operation.*',
 'status.operation.calibrating.*',
 'status.operation.instrument.digio.*',
 'status.operation.instrument.digio.trigger_overrun.*',
 'status.operation.instrument.lan.*',
 'status.operation.instrument.lan.trigger_overrun.*',
 'status.operation.instrument.smua.*',
 'status.operation.instrument.smua.trigger_overrrun.*',
 'status.operation.instrument.smub.*',
 'status.operation.instrument.smub.trigger_overrrun.*',
 'status.operation.instrument.trigger_blender.*',
 'status.operation.instrument.trigger_blender.trigger_overrun.*',
 'status.operation.instrument.trigger_timer.*',
 'status.operation.instrument.trigger_timer.trigger_overrun.*',
 'status.operation.instrument.tsplink.*',
 'status.operation.instrument.tsplink.trigger_overrun.*',
 'status.operation.measuring.*',
 'status.operation.remote.*',
 'status.operation.sweeping.*',
 'status.operation.trigger_overrun.*',
 'status.operation.user.*',
 'status.questionable.*',
 'status.questionable.calibration.*',
 'status.questionable.instrument.*',
 'status.questionable.instrument.smua.*',
 'status.questionable.instrument.smub.*',
 'status.questionable.over_temperature.*',
 'status.questionable.unstable_output.*',
 'status.request_enable',
 'status.request_event',
 'status.reset',
 'status.standard.*',
 'status.system.*',
 'status.system2.*',
 'status.system3.*',
 'status.system4.*',
 'status.system5.*',
 'timer.measure.t',
 'timer.reset',
 'trigger.EVENT_ID',
 'trigger.blender.clear',
 'trigger.blender.reset',
 'trigger.blender.wait',
 'trigger.blender[N].EVENT_ID',
 'trigger.blender[N].orenable',
 'trigger.blender[N].overrun',
 'trigger.blender[N].stimulus[M]',
 'trigger.clear',
 'trigger.timer.clear',
 'trigger.timer.reset',
 'trigger.timer.wait',
 'trigger.timer[N].EVENT_ID',
 'trigger.timer[N].count',
 'trigger.timer[N].delay',
 'trigger.timer[N].delaylist',
 'trigger.timer[N].overrun',
 'trigger.timer[N].passthrough',
 'trigger.timer[N].stimulus',
 'trigger.wait',
 'tsplink.group',
 'tsplink.master',
 'tsplink.readbit',
 'tsplink.readport',
 'tsplink.reset',
 'tsplink.state',
 'tsplink.trigger.assert',
 'tsplink.trigger.clear',
 'tsplink.trigger.release',
 'tsplink.trigger.reset',
 'tsplink.trigger.wait',
 'tsplink.trigger[N].EVENT_ID',
 'tsplink.trigger[N].mode',
 'tsplink.trigger[N].overrun',
 'tsplink.trigger[N].pulsewidth',
 'tsplink.trigger[N].stimulus',
 'tsplink.writebit',
 'tsplink.writeport',
 'tsplink.writeprotect',
 'tspnet.clear',
 'tspnet.connect',
 'tspnet.disconnect',
 'tspnet.execute',
 'tspnet.idn',
 'tspnet.read',
 'tspnet.readavailable',
 'tspnet.reset',
 'tspnet.termination',
 'tspnet.timeout',
 'tspnet.tsp.abort',
 'tspnet.tsp.abortonconnect',
 'tspnet.tsp.rbtablecopy',
 'tspnet.tsp.runscript',
 'tspnet.write',
 'userstring.add',
 'userstring.catalog',
 'userstring.delete',
 'userstring.get',
 'waitcomplete',
 ]

PROPERTY_LISTS = [
        'trigger.blender[1].stimulus',
        'trigger.blender[2].stimulus',
        'smua.nvbuffer1.readings',
        'smua.nvbuffer2.readings',
        'smub.nvbuffer1.readings',
        'smub.nvbuffer2.readings',
        'readings',
        ]

properties = set(PROPERTIES)
functions = set(FUNCTIONS)
constants = set(CONSTANTS)
classes = set(CLASSES)

check0 = bool(properties & classes)
check1 = bool(functions & classes)
check2 = bool(constants & classes)

if check0 or check1 or check2:
    raise RuntimeError(WARNING)



# -*- coding: utf-8 -*-
"""
MIT License
===========
Copyright (c) 2009- keithley2600 Project Contributors
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:
The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

from keithley2600.sweep_data import ResultTable, TransistorSweepData, IVSweepData
from keithley2600.keithley_driver import Keithley2600Factory as Keithley2600
from keithley2600.keithley_driver import log_to_screen



# -*- coding: utf-8 -*-
#
# Copyright © keithley2600 Project Contributors
# Licensed under the terms of the MIT License
# (see keithley2600/__init__.py for details)

"""
Core driver with the low level functions.
"""

# system imports
from __future__ import absolute_import, division, print_function
import sys
import visa
import logging
import threading
import numpy as np
import time

# local import
from keithley2600.keithley_doc import (CONSTANTS, FUNCTIONS, PROPERTIES,
                                       CLASSES, PROPERTY_LISTS)
from keithley2600.sweep_data import TransistorSweepData

PY2 = sys.version[0] == '2'
logger = logging.getLogger(__name__)

if not PY2:
    basestring = str  # in Python 3


def log_to_screen(level=logging.DEBUG):
    log_to_stream(None, level)  # sys.stderr by default


def log_to_stream(stream_output, level=logging.DEBUG):
    logger.setLevel(level)
    ch = logging.StreamHandler(stream_output)
    ch.setLevel(level)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    ch.setFormatter(formatter)

    logger.addHandler(ch)


class MagicPropertyList(object):
    """Mimics a Keithley TSP property list

    Class which mimics a Keithley TSP property list and can be dynamically
    created. It forwards all calls to the _read method of the parent class and
    assignments to the _write method. Arbitrary values can be assigned, as
    long as _write can handle them.

    This class is designed to look like a  Keithley TSP "attribute" list,
    forward function calls to the Keithley, and return the results.

    """

    def __init__(self, name, parent):
        if not isinstance(name, basestring):
            raise ValueError('First argument must be of type str.')
        self._name = name
        self._parent = parent

    def __getitem__(self, i):
        """Gets i-th item: query item from parent class

        Args:
            i: An integer item number

        Returns:
            Result from _query call of parent class.
        """
        new_name = '%s[%s]' % (self._name, i)
        return self._query(new_name)

    def __setitem__(self, i, value):
        """Sets i-th item: set item at parent class

        Args:
            i: An integer item number
            value: An input object that can be accepted by parent class.

        Returns:
            None.
        """
        value = self._convert_input(value)
        new_name = '%s[%s] = %s' % (self._name, i, value)
        self._write(new_name)

    def __iter__(self):
        return self

    def _write(self, value):
        """Forward _write calls to parent class."""
        self._parent._write(value)

    def _query(self, value):
        """Forward _query calls to parent class."""
        return self._parent._query(value)

    def _convert_input(self, value):
        """Forward _convert_input calls to parent class."""
        try:
            return self._parent._convert_input(value)
        except AttributeError:
            return value

    def getdoc(self):
        """Prevent pydoc from trying to document this class. This could
        conflict with on-demand creation of attributes."""
        pass


class MagicFunction(object):
    """Mimics a Keithley TSP function

    Class which mimics a function and can be dynamically created. It forwards
    all calls to the _query method of the parent class and returns the result
    from _query. Calls accept arbitrary arguments, as long as _query can
    handle them.

    This class is designed to look like a Keithley TSP function, forward
    function calls to the Keithley, and return the results.

    """

    def __init__(self, name, parent):
        if not isinstance(name, basestring):
            raise ValueError('First argument must be of type str.')
        self._name = name
        self._parent = parent

    def __call__(self, *args, **kwargs):
        """Pass on calls to `self.prent._write`, store result in variable.
        Querying results from function calls directly may result in
        a VisaIOError if the function does not return anything."""

        # convert incompatible arguments, return all arguments as tuple
        args = tuple(self._parent._convert_input(a) for a in args)
        # remove outside brackets and all quotation marks
        args_string = str(args).strip("(),").replace("'", "")
        # pass on calls to self._write as string representing function call
        self._parent._write('result = %s(%s)' % (self._name, args_string))
        # query for result in second call
        return self._parent._query('result')


class MagicClass(object):
    """Mimics a TSP command group

    Class which dynamically creates new attributes on access. These can be
    functions, properties, or other classes.

    MagicClass need the strings in FUNCTIONS and PROPERTIES to determine if the
    accessed attribute should behave like a function or property. Otherwise, it
    is assumed to be a new class.

    Attribute setters and getters are forwarded to _write and _query functions
    from the parent class. New functions are created as instances of
    MagicFunction, new classes are created as instances of MagicClass.

    MagicClass is designed to mimic a Keithley TSP command group with
    functions, attributes, and subordinate command groups.

    USAGE:
        inst = MagicClass('keithley')
        inst.reset() - Dynamically creates a new attribute 'reset' as an instance
                       of MagicFunction, then calls it.
        inst.beeper  - Dynamically creates new attribute 'beeper' and sets it to
                       a new MagicClass instance.
        inst.beeper.enable - Fakes the property 'enable' of 'beeper'
                             with _write as setter and _query as getter.

    """

    _name = ''
    _parent = None

    def __init__(self, name, parent=None):
        if not isinstance(name, basestring):
            raise ValueError('First argument must be of type str.')
        self._name = name
        self._parent = parent

    def __getattr__(self, attr_name):
        """Custom getter

        Get attributes as usual if they exist. Otherwise, fall back to
        `self.__get_global_handler`.
        """
        try:
            try:
                # check if attribute already exists. return attr if yes.
                return object.__getattr__(self, attr_name)
            except AttributeError:
                # check if key already exists. return value if yes.
                return self.__dict__[attr_name]
        except KeyError:
            # handle if not
            return self.__get_global_handler(attr_name)

    def __get_global_handler(self, attr_name):
        """Custom getter

        Create attribute as MagicClass, MagicFunction or MagicPropertyList
        instance if it is an expected Keithley TSP command group, function or
        property list. Query and return value if attribute corresponds to a
        Keithley TSP constant. Otherwise raise standard AttributeError.

        Args:
            attr_name: Attribute name.

        Returns:
            Instance or MagicClass, MagicFunction or MagicPropertyList.

        Raises:
            AttributeError if attribute is not expected.
        """

        # create callable sub-class for new attr
        new_name = '%s.%s' % (self._name, attr_name)
        new_name = new_name.strip('.')

        if attr_name in FUNCTIONS:
            handler = MagicFunction(new_name, parent=self)
            self.__dict__[new_name] = handler

        elif attr_name in PROPERTY_LISTS:
            handler = MagicPropertyList(new_name, parent=self)

        elif attr_name in PROPERTIES or attr_name in CONSTANTS:
            if new_name in PROPERTY_LISTS:
                handler = MagicPropertyList(new_name, parent=self)
            else:
                handler = self._query(new_name)

        elif attr_name in CLASSES:
            handler = MagicClass(new_name, parent=self)
            self.__dict__[new_name] = handler

        else:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self), attr_name)
                )

        return handler

    def __setattr__(self, attr_name, value):
        """Custom setter

        Forward setting commands to `self._write` for expected Keithley TSP
        attributes. Otherwise use default setter.

        Args:
            attr_name: Attribute name.
            value: Value to set.

        Returns:
            None.

        Raises:
            ValueError if trying to write a value to read-only Keithley
            attributes.
        """
        if attr_name in PROPERTIES:
            value = self._convert_input(value)
            self._write('%s.%s = %s' % (self._name, attr_name, value))
        elif attr_name in CONSTANTS:
            raise ValueError('%s.%s is read-only.' % (self._name, attr_name))
        else:
            object.__setattr__(self, attr_name, value)
            self.__dict__[attr_name] = value

    def _write(self, value):
        """Forward _write calls to parent class."""
        self._parent._write(value)

    def _query(self, value):
        """Forward _query calls to parent class."""
        return self._parent._query(value)

    def _convert_input(self, value):
        """Forward _convert_input calls to parent class."""
        try:
            return self._parent._convert_input(value)
        except AttributeError:
            return value

    def __getitem__(self, i):
        """Return new MagicClass instance for every item."""
        new_name = '%s[%s]' % (self._name, i)
        new_class = MagicClass(new_name, parent=self)
        return new_class

    def __iter__(self):
        return self

    def getdoc(self):
        """Prevent pydoc from trying to document this class. This could
        conflict with on-demand creation of attributes."""
        pass


class KeithleyIOError(Exception):
    pass


class Keithley2600Base(MagicClass):
    """Keithley2600 driver

    Keithley driver for base functions. It replicates the functionality and
    syntax from the Keithley TSP commands, which have a syntax similar to
    python. Attributes are created on-access if they correspond to Keithley TSP
    type commands.

    Warning:
        There are currently no checks for allowed arguments in the base
        commands. See the Keithley 2600 reference manual for all available
        commands and arguments. Almost all remotely accessible commands can be
        used with this driver. NOT SUPPORTED ARE:
             * `tspnet.execute()` - conflicts with Python's execute command
             * `lan.trigger[N].connected` - conflicts with the connected attribute of
                Keithley2600Base
             * All Keithley IV sweep commands. We implement our own in the
               Keithley2600 class.

    Example:
        >>> keithley = Keithley2600Base('TCPIP0::192.168.2.121::INSTR')
        >>> keithley.smua.measure.v()  # measures the smuA voltage
        >>> keithley.smua.source.levelv = -40  # applies -40V to smuA

    Documentation:
        See the Keithley 2600 reference manual for all available commands and
        arguments.

    Attributes:
        _lock (threading.RLock): Lock to prevent simultaneous calls to
            the Keithley.
        connection (visa resource): Attribute holding reference to the actual
            connection.
        connected (bool): Attribute to hold info if connected.
    """

    _lock = threading.RLock()
    connection = None
    connected = False
    busy = False

    # input types that will be accepted as TSP lists by keithley
    TO_TSP_LIST = (list, np.ndarray, tuple, set)
    # maximum length of lists send to keithley
    CHUNK_SIZE = 50

    def __init__(self, visa_address, visa_library='@py', **kwargs):
        """Initializes driver, connects to Keithley

        Args:
            visa_address: Visa address of Keithley containing connection type
                and system address, e.g., "TCPIP0::192.168.2.121::INSTR". See
                NI-VISA for PyVisa for documentation
            visa_library: Visa backend used by PyVisa. Can be an empty string
                for NI-VISA backend, a path to the visa library, or "@py" for
                the py-visa-py backend.
        """

        MagicClass.__init__(self, name='', parent=self)
        self._name = ''  # visa_address will

        self.abort_event = threading.Event()

        self.visa_address = visa_address
        self.visa_library = visa_library

        # open visa resource manager with selected library / backend
        self.rm = visa.ResourceManager(self.visa_library)
        # connect to keithley
        self.connect(**kwargs)

    def __repr__(self):
        return '<%s(%s)>' % (type(self).__name__, self.visa_address)

# =============================================================================
# Connect to keithley
# =============================================================================

    def connect(self, **kwargs):
        """
        Connects to Keithley and opens pyvisa API.
        """
        connection_error = OSError if PY2 else ConnectionError
        # noinspection PyBroadException
        try:
            self.connection = self.rm.open_resource(self.visa_address, **kwargs)
            self.connection.read_termination = '\n'
            self.connected = True
            logger.debug('Connected to Keithley at %s.' % self.visa_address)
        except ValueError:
            self.connection = None
            self.connected = False
            raise
        except connection_error:
            logger.info('Connection error. Please check that ' +
                        'no other program is connected.')
            self.connection = None
            self.connected = False
        except AttributeError:
            logger.info('Invalid VISA address %s.' % self.visa_address)
            self.connection = None
            self.connected = False
        except Exception:
            logger.info('Could not connect to Keithley at %s.' % self.visa_address)
            self.connection = None
            self.connected = False

    def disconnect(self):
        """ Disconnect from Keithley """
        if self.connection:
            try:
                self.connection.close()
                self.connection = None
                self.connected = False
                del self.connection
                logger.debug('Disconnected from Keithley at %s.' % self.visa_address)
            except AttributeError:
                self.connected = False
                pass

# =============================================================================
# Define I/O
# =============================================================================

    def _write(self, value):
        """
        Writes text to Keithley. Input must be a string.
        """
        logger.debug('write: %s' % value)

        if self.connection:
            self.connection.write(value)
        else:
            raise KeithleyIOError(
                'No connection to keithley present. Try to call connect().')

    def _query(self, value):
        """
        Queries and expects response from Keithley. Input must be a string.
        """
        logger.debug('write: print(%s)' % value)

        if self.connection:
            with self._lock:
                r = self.connection.query('print(%s)' % value)
                logger.debug('read: %s' % r)

            return self.parse_response(r)
        else:
            raise KeithleyIOError(
                'No connection to keithley present. Try to call connect().')

    @staticmethod
    def parse_response(string):
        try:
            r = float(string)
        except ValueError:
            if string == 'nil':
                r = None
            elif string == 'true':
                r = True
            elif string == 'false':
                r = False
            else:
                r = string

        return r

    def _convert_input(self, value):
        """ Convert bool to lower case string and list / tuples to comma
        delimited string enclosed by curly brackets."""
        if isinstance(value, bool):
            # convert bool True to string 'true'
            value = str(value).lower()
        elif isinstance(value, self.TO_TSP_LIST):
            # convert some iterables to a TSP type list '{1,2,3,4}'
            value = '{%s}' % ', '.join(map(str, value))
        elif isinstance(value, MagicClass):
            # convert keithley object to string with its name
            value = value._name

        return value


class Keithley2600(Keithley2600Base):
    """Keithley2600 driver with high level functionality

    Keithley driver with access to base functions and higher level functions
    such as IV measurements, transfer and output curves, etc. Base command
    replicate the functionality and syntax from the Keithley TSP functions,
    which have a syntax similar to python.

    Warning:
        There are currently no checks for allowed arguments in the base
        commands. See the Keithley 2600 reference manual for all available
        commands and arguments. Almost all remotely accessible commands can be
        used with this driver. NOT SUPPORTED ARE:
             * tspnet.execute() # conflicts with Python's execute command
             * All Keithley IV sweep commands. We implement our own here.

    Example:
        Base commands from keithley TSP:

        >>> k = Keithley2600('TCPIP0::192.168.2.121::INSTR')
        >>> volts = k.smua.measure.v()  # measures and returns the smuA voltage
        >>> k.smua.source.levelv = -40  # sets source level of smuA
        >>> k.smua.nvbuffer1.clear()  # clears nvbuffer1 of smuA

        New mid-level commands:

        >>> data = k.readBuffer(k.smua.nvbuffer1)
        >>> k.setIntegrationTime(k.smua, 0.001) # in sec

        >>> k.applyVoltage(k.smua, -60) # applies -60V to smuA
        >>> k.applyCurrent(k.smub, 0.1) # sources 0.1A from smuB
        >>> k.rampToVoltage(k.smua, 10, delay=0.1, step_size=1)

        >>> # voltage sweeps, single and dual SMU
        >>> k.voltageSweepSingleSMU(smu=k.smua, smu_sweeplist=list(range(0, 61)),
        ...                         t_int=0.1, delay=-1, pulsed=False)
        >>> k.voltageSweepDualSMU(smu1=k.smua, smu2=k.smub,
        ...                       smu1_sweeplist=list(range(0, 61)),
        ...                       smu2_sweeplist=list(range(0, 61)),
        ...                       t_int=0.1, delay=-1, pulsed=False)

        New high-level commands:

        >>> data1 = k.outputMeasurement(...) # records output curve
        >>> data2 = k.transferMeasurement(...) # records transfer curve

    Attributes:
        SMU_LIST (list): List containing strings of all smu column_names.
    """

    SMU_LIST = ['smua', 'smub']

    def __init__(self, visa_address, visa_library='@py', **kwargs):
        Keithley2600Base.__init__(self, visa_address, visa_library, **kwargs)

    def __repr__(self):
        return '<%s(%s)>' % (type(self).__name__, self.visa_address)

    def _check_smu(self, smu):
        """Check if selected smu is indeed present."""
        assert smu._name.split('.')[-1] in self.SMU_LIST

    @staticmethod
    def _get_smu_string(smu):
        return smu._name.split('.')[-1]

# =============================================================================
# Define lower level control functions
# =============================================================================

    @staticmethod
    def readBuffer(buffer):
        """
        Reads buffer values and returns them as a list. This can be done more
        quickly by "readings = buffer.readings" but such a call may fail due to
        I/O limitations of the keithley if the readings list is too long.

        Args:
            buffer: A keithley buffer instance.

        Returns:
            A list with buffer readings.
        """
        list_out = []
        for i in range(0, int(buffer.n)):
            list_out.append(buffer.readings[i+1])

        return list_out

    def clearBuffer(self, smu):
        """
        Clears buffer of a given smu. This function has been deprecated.
        """

        raise DeprecationWarning(
            "'clearBuffer()' has been deprecated. Please use 'buffer.clear()' and " +
            "'buffer.clearcache()' instead, where 'buffer' is a Keithley2600 buffer " +
            "instance such as 'k.smua.nvbuffer1'.")

    def setIntegrationTime(self, smu, t_int):
        """
        Sets the integration time of SMU for measurements in sec.

        Args:
            smu: A keithley smu instance.
            t_int (float): Integration time in sec. Value must be between 0.001
                and 25 power line cycles.
        Raises:
            ValueError for too short or too long integration times.
        """

        self._check_smu(smu)

        # determine number of power-line-cycles used for integration
        freq = self.localnode.linefreq
        nplc = t_int * freq

        if nplc < 0.001 or nplc > 25:
            raise ValueError('Integration time must be between 0.001 and 25 ' +
                             'power line cycles of 1/(%s Hz).' % freq)
        smu.measure.nplc = nplc

    def applyVoltage(self, smu, voltage):
        """
        Turns on the specified SMU and applies a voltage.
        """

        self._check_smu(smu)

        smu.source.output = smu.OUTPUT_ON
        smu.source.levelv = voltage

    def applyCurrent(self, smu, curr):
        """
        Turns on the specified SMU and sources a current.
        """
        self._check_smu(smu)

        smu.source.leveli = curr
        smu.source.output = smu.OUTPUT_ON

    def rampToVoltage(self, smu, target_volt, delay=0.1, step_size=1):
        """
        Ramps up the voltage of the specified SMU. Beeps when done.

        Args:
            smu: SMU to ramp.
            target_volt (float): Target gate voltage.
            step_size (float): Size of the voltage ramp steps in Volts.
            delay (float): Delay between steps in sec.
        """

        self._check_smu(smu)

        smu.source.output = smu.OUTPUT_ON

        # get current voltage
        vcurr = smu.source.levelv
        if vcurr == target_volt:
            return

        self.display.smua.measure.func = self.display.MEASURE_DCVOLTS
        self.display.smub.measure.func = self.display.MEASURE_DCVOLTS

        step = np.sign(target_volt - vcurr) * abs(step_size)

        for v in np.arange(vcurr, target_volt + step, step):
            smu.source.levelv = v
            smu.measure.v()
            time.sleep(delay)

        target_volt = smu.measure.v()
        logger.info('Gate voltage set to Vg = %s V.' % round(target_volt))

        self.beeper.beep(0.3, 2400)

    def voltageSweepSingleSMU(self, smu, smu_sweeplist, t_int, delay, pulsed):
        """
        Sweeps voltage at one SMU. Measures and returns current and voltage during sweep.

        Args:
            smu (keithley smu object): 1st SMU to be swept.
            smu_sweeplist (list): Voltages to sweep through, must be a list,
                tuple, or numpy array.
            t_int (float): Integration time per data point.
            delay (float): Settling delay before measurement.
            pulsed (bool): True or False for pulsed or continuous sweep.

        Returns:
            v_smu (list): Voltages measurement during the sweep in Volts.
            i_smu (list): Currents measurement during the sweep in Amperes.
        """

        # input checks
        self._check_smu(smu)

        # set state to busy
        self.busy = True
        # Define lists containing results. If we abort early, we have something to return.
        v_smu, i_smu = [], []

        if self.abort_event.is_set():
            self.busy = False
            return v_smu, i_smu

        # setup smu to sweep through list on trigger
        # send sweep_list over in chunks if too long
        if len(smu_sweeplist) > self.CHUNK_SIZE:
            self._write('mylist = {}')
            for num in smu_sweeplist:
                self._write('table.insert(mylist, %s)' % num)
            smu.trigger.source.listv('mylist')
        else:
            smu.trigger.source.listv(smu_sweeplist)

        smu.trigger.source.action = smu.ENABLE

        # CONFIGURE INTEGRATION TIME FOR EACH MEASUREMENT
        self.setIntegrationTime(smu, t_int)

        # CONFIGURE SETTLING TIME FOR GATE VOLTAGE, I-LIMIT, ETC...
        smu.measure.delay = delay
        smu.measure.autorangei = smu.AUTORANGE_ON

        # smu.trigger.source.limiti = 0.1

        smu.source.func = smu.OUTPUT_DCVOLTS

        # 2-wire measurement (use SENSE_REMOTE for 4-wire)
        # smu.sense = smu.SENSE_LOCAL

        # clears SMU buffers
        smu.nvbuffer1.clear()
        smu.nvbuffer2.clear()

        smu.nvbuffer1.clearcache()
        smu.nvbuffer2.clearcache()

        # display current values during measurement
        self.display.smua.measure.func = self.display.MEASURE_DCAMPS
        self.display.smub.measure.func = self.display.MEASURE_DCAMPS

        # SETUP TRIGGER ARM AND COUNTS
        # trigger count = number of data points in measurement
        # arm count = number of times the measurement is repeated (set to 1)

        npts = len(smu_sweeplist)
        smu.trigger.count = npts

        # SET THE MEASUREMENT TRIGGER ON BOTH SMU'S
        # Set measurement to trigger once a change in the gate value on
        # sweep smu is complete, i.e., a measurement will occur
        # after the voltage is stepped.
        # Both channels should be set to trigger on the sweep smu event
        # so the measurements occur at the same time.

        # enable smu
        smu.trigger.measure.action = smu.ENABLE

        # measure current and voltage on trigger, store in buffer of smu
        smu.trigger.measure.iv(smu.nvbuffer1, smu.nvbuffer2)

        # initiate measure trigger when source is complete
        smu.trigger.measure.stimulus = smu.trigger.SOURCE_COMPLETE_EVENT_ID

        # SET THE ENDPULSE ACTION TO HOLD
        # Options are SOURCE_HOLD AND SOURCE_IDLE, hold maintains same voltage
        # throughout step in sweep (typical IV sweep behavior). idle will allow
        # pulsed IV sweeps.

        if pulsed:
            end_pulse_action = 0  # SOURCE_IDLE
        elif not pulsed:
            end_pulse_action = 1  # SOURCE_HOLD
        else:
            raise TypeError("'pulsed' must be of type 'bool'.")

        smu.trigger.endpulse.action = end_pulse_action

        # SET THE ENDSWEEP ACTION TO HOLD IF NOT PULSED
        # Output voltage will be held after sweep is done!

        smu.trigger.endsweep.action = end_pulse_action

        # SET THE EVENT TO TRIGGER THE SMU'S TO THE ARM LAYER
        # A typical measurement goes from idle -> arm -> trigger.
        # The 'trigger.event_id' option sets the transition arm -> trigger
        # to occur after sending *trg to the instrument.

        smu.trigger.arm.stimulus = self.trigger.EVENT_ID

        # Prepare an event blender (blender #1) that triggers when
        # the smua enters the trigger layer or reaches the end of a
        # single trigger layer cycle.

        # triggers when either of the stimuli are true ('or enable')
        self.trigger.blender[1].orenable = True
        self.trigger.blender[1].stimulus[1] = smu.trigger.ARMED_EVENT_ID
        self.trigger.blender[1].stimulus[2] = smu.trigger.PULSE_COMPLETE_EVENT_ID

        # SET THE smu SOURCE STIMULUS TO BE EVENT BLENDER #1
        # A source measure cycle within the trigger layer will occur when
        # either the trigger layer is entered (termed 'armed event') for the
        # first time or a single cycle of the trigger layer is complete (termed
        # 'pulse complete event').

        smu.trigger.source.stimulus = self.trigger.blender[1].EVENT_ID

        # PREPARE AN EVENT BLENDER (blender #2) THAT TRIGGERS WHEN BOTH SMU'S
        # HAVE COMPLETED A MEASUREMENT.
        # This is needed to prevent the next source measure cycle from occurring
        # before the measurement on both channels is complete.

        self.trigger.blender[2].orenable = True  # triggers when both stimuli are true
        self.trigger.blender[2].stimulus[1] = smu.trigger.MEASURE_COMPLETE_EVENT_ID

        # SET THE SMU ENDPULSE STIMULUS TO BE EVENT BLENDER #2
        smu.trigger.endpulse.stimulus = self.trigger.blender[2].EVENT_ID

        # TURN ON smu
        smu.source.output = smu.OUTPUT_ON

        # INITIATE MEASUREMENT
        # prepare SMUs to wait for trigger
        smu.trigger.initiate()

        # send trigger
        self._write('*trg')

        # CHECK STATUS BUFFER FOR MEASUREMENT TO FINISH
        # Possible return values:
        # 6 = smua and smub sweeping
        # 4 = only smub sweeping
        # 2 = only smua sweeping
        # 0 = neither smu sweeping

        status = 0
        while status == 0:  # while loop that runs until the sweep begins
            status = self.status.operation.sweeping.condition
            time.sleep(0.1)

        while status > 0:  # while loop that runs until the sweep ends
            status = self.status.operation.sweeping.condition
            time.sleep(0.1)

        # EXTRACT DATA FROM SMU BUFFERS
        i_smu = self.readBuffer(smu.nvbuffer1)
        v_smu = self.readBuffer(smu.nvbuffer2)

        smu.nvbuffer1.clear()
        smu.nvbuffer2.clear()

        smu.nvbuffer1.clearcache()
        smu.nvbuffer2.clearcache()

        self.busy = False

        return v_smu, i_smu

    def voltageSweepDualSMU(self, smu1, smu2, smu1_sweeplist, smu2_sweeplist, t_int,
                            delay, pulsed):
        """
        Sweeps voltages at two SMUs. Measures and returns current and voltage during
        sweep.

        Args:
            smu1 (keithley smu object): 1st SMU to be swept.
            smu2 (keithley smu object): 2nd SMU to be swept.
            smu1_sweeplist: List of voltages to sweep at smu1 (can be a numpy
                 array, list or tuple).
            smu2_sweeplist: List of voltages to sweep at smu2 (can be a numpy
                array, list or tuple).
            t_int (float): Integration time per data point (float), must be
                between 0.001 to 25 times the power line frequency
            delay (float): Settling delay before measurement.
            pulsed (float): Continuous or pulsed sweep.

        Returns:
            v_smu1 (list): Voltages measurement during the sweep in Volts at
                the first smu.
            i_smu1 (list): Currents measurement during the sweep in Amperes at
                the first smu.
            v_smu2 (list): Voltages measurement during the sweep in Volts at
                the second smu.
            i_smu2 (list): Currents measurement during the sweep in Amperes at
                the second smu.
        """

        # input checks
        self._check_smu(smu1)
        self._check_smu(smu2)

        assert len(smu1_sweeplist) == len(smu2_sweeplist)

        # set state to busy
        self.busy = True
        # Define lists containing results. If we abort early, we have something to return.
        v_smu1, i_smu1, v_smu2, i_smu2 = [], [], [], []

        if self.abort_event.is_set():
            self.busy = False
            return v_smu1, i_smu1, v_smu2, i_smu2

        # Setup smua/smub for sweep measurement.

        # setup smu1 and smu2 to sweep through lists on trigger
        # send sweep_list over in chunks if too long
        if len(smu1_sweeplist) > self.CHUNK_SIZE:
            self._write('mylist = {}')
            for num in smu1_sweeplist:
                self._write('table.insert(mylist, %s)' % num)
            smu1.trigger.source.listv('mylist')
        else:
            smu1.trigger.source.listv(smu1_sweeplist)

        if len(smu2_sweeplist) > self.CHUNK_SIZE:
            self._write('mylist = {}')
            for num in smu2_sweeplist:
                self._write('table.insert(mylist, %s)' % num)
            smu2.trigger.source.listv('mylist')
        else:
            smu2.trigger.source.listv(smu2_sweeplist)

        smu1.trigger.source.action = smu1.ENABLE
        smu2.trigger.source.action = smu2.ENABLE

        # CONFIGURE INTEGRATION TIME FOR EACH MEASUREMENT
        self.setIntegrationTime(smu1, t_int)
        self.setIntegrationTime(smu2, t_int)

        # CONFIGURE SETTLING TIME FOR GATE VOLTAGE, I-LIMIT, ETC...
        smu1.measure.delay = delay
        smu2.measure.delay = delay

        smu1.measure.autorangei = smu1.AUTORANGE_ON
        smu2.measure.autorangei = smu2.AUTORANGE_ON

        # smu1.trigger.source.limiti = 0.1
        # smu2.trigger.source.limiti = 0.1

        smu1.source.func = smu1.OUTPUT_DCVOLTS
        smu2.source.func = smu2.OUTPUT_DCVOLTS

        # 2-wire measurement (use SENSE_REMOTE for 4-wire)
        # smu1.sense = smu1.SENSE_LOCAL
        # smu2.sense = smu2.SENSE_LOCAL

        # CLEAR BUFFERS
        for smu in [smu1, smu2]:
            smu.nvbuffer1.clear()
            smu.nvbuffer2.clear()
            smu.nvbuffer1.clearcache()
            smu.nvbuffer2.clearcache()

        # display current values during measurement
        self.display.smua.measure.func = self.display.MEASURE_DCAMPS
        self.display.smub.measure.func = self.display.MEASURE_DCAMPS

        # SETUP TRIGGER ARM AND COUNTS
        # trigger count = number of data points in measurement
        # arm count = number of times the measurement is repeated (set to 1)

        npts = len(smu1_sweeplist)

        smu1.trigger.count = npts
        smu2.trigger.count = npts

        # SET THE MEASUREMENT TRIGGER ON BOTH SMU'S
        # Set measurement to trigger once a change in the gate value on
        # sweep smu is complete, i.e., a measurement will occur
        # after the voltage is stepped.
        # Both channels should be set to trigger on the sweep smu event
        # so the measurements occur at the same time.

        # enable smu
        smu1.trigger.measure.action = smu1.ENABLE
        smu2.trigger.measure.action = smu2.ENABLE

        # measure current and voltage on trigger, store in buffer of smu
        smu1.trigger.measure.iv(smu1.nvbuffer1, smu1.nvbuffer2)
        smu2.trigger.measure.iv(smu2.nvbuffer1, smu2.nvbuffer2)

        # initiate measure trigger when source is complete
        smu1.trigger.measure.stimulus = smu1.trigger.SOURCE_COMPLETE_EVENT_ID
        smu2.trigger.measure.stimulus = smu1.trigger.SOURCE_COMPLETE_EVENT_ID

        # SET THE ENDPULSE ACTION TO HOLD
        # Options are SOURCE_HOLD AND SOURCE_IDLE, hold maintains same voltage
        # throughout step in sweep (typical IV sweep behavior). idle will allow
        # pulsed IV sweeps.

        if pulsed:
            end_pulse_action = 0  # SOURCE_IDLE
        elif not pulsed:
            end_pulse_action = 1  # SOURCE_HOLD
        else:
            raise TypeError("'pulsed' must be of type 'bool'.")

        smu1.trigger.endpulse.action = end_pulse_action
        smu2.trigger.endpulse.action = end_pulse_action

        # SET THE ENDSWEEP ACTION TO HOLD IF NOT PULSED
        # Output voltage will be held after sweep is done!

        smu1.trigger.endsweep.action = end_pulse_action
        smu2.trigger.endsweep.action = end_pulse_action

        # SET THE EVENT TO TRIGGER THE SMU'S TO THE ARM LAYER
        # A typical measurement goes from idle -> arm -> trigger.
        # The 'trigger.event_id' option sets the transition arm -> trigger
        # to occur after sending *trg to the instrument.

        smu1.trigger.arm.stimulus = self.trigger.EVENT_ID

        # Prepare an event blender (blender #1) that triggers when
        # the smua enters the trigger layer or reaches the end of a
        # single trigger layer cycle.

        # triggers when either of the stimuli are true ('or enable')
        self.trigger.blender[1].orenable = True
        self.trigger.blender[1].stimulus[1] = smu1.trigger.ARMED_EVENT_ID
        self.trigger.blender[1].stimulus[2] = smu1.trigger.PULSE_COMPLETE_EVENT_ID

        # SET THE smu1 SOURCE STIMULUS TO BE EVENT BLENDER #1
        # A source measure cycle within the trigger layer will occur when
        # either the trigger layer is entered (termed 'armed event') for the
        # first time or a single cycle of the trigger layer is complete (termed
        # 'pulse complete event').

        smu1.trigger.source.stimulus = self.trigger.blender[1].EVENT_ID

        # PREPARE AN EVENT BLENDER (blender #2) THAT TRIGGERS WHEN BOTH SMU'S
        # HAVE COMPLETED A MEASUREMENT.
        # This is needed to prevent the next source measure cycle from occurring
        # before the measurement on both channels is complete.

        self.trigger.blender[2].orenable = False  # triggers when both stimuli are true
        self.trigger.blender[2].stimulus[1] = smu1.trigger.MEASURE_COMPLETE_EVENT_ID
        self.trigger.blender[2].stimulus[2] = smu2.trigger.MEASURE_COMPLETE_EVENT_ID

        # SET THE smu1 ENDPULSE STIMULUS TO BE EVENT BLENDER #2
        smu1.trigger.endpulse.stimulus = self.trigger.blender[2].EVENT_ID

        # TURN ON smu1 AND smu2
        smu1.source.output = smu1.OUTPUT_ON
        smu2.source.output = smu2.OUTPUT_ON

        # INITIATE MEASUREMENT
        # prepare SMUs to wait for trigger
        smu1.trigger.initiate()
        smu2.trigger.initiate()
        # send trigger
        self._write('*trg')

        # CHECK STATUS BUFFER FOR MEASUREMENT TO FINISH
        # Possible return values:
        # 6 = smua and smub sweeping
        # 4 = only smub sweeping
        # 2 = only smua sweeping
        # 0 = neither smu sweeping

        status = 0
        while status == 0:  # while loop that runs until the sweep begins
            status = self.status.operation.sweeping.condition
            time.sleep(0.1)

        while status > 0:  # while loop that runs until the sweep ends
            status = self.status.operation.sweeping.condition
            time.sleep(0.1)

        # EXTRACT DATA FROM SMU BUFFERS
        i_smu1 = self.readBuffer(smu1.nvbuffer1)
        v_smu1 = self.readBuffer(smu1.nvbuffer2)
        i_smu2 = self.readBuffer(smu2.nvbuffer1)
        v_smu2 = self.readBuffer(smu2.nvbuffer2)

        # CLEAR BUFFERS
        for smu in [smu1, smu2]:
            smu.nvbuffer1.clear()
            smu.nvbuffer2.clear()
            smu.nvbuffer1.clearcache()
            smu.nvbuffer2.clearcache()

        self.busy = False

        return v_smu1, i_smu1, v_smu2, i_smu2

# =============================================================================
# Define higher level control functions
# =============================================================================

    def transferMeasurement(self, smu_gate, smu_drain, vg_start, vg_stop,
                            vg_step, vd_list, t_int, delay, pulsed):
        """
        Records a transfer curve and saves the results in a TransistorSweepData
        instance.

        Args:
            smu_gate: SMU attached to gate electrode of FET for transfer
                measurement (keithley smu object).
            smu_drain: SMU attached to drain electrode of FET for transfer
                measurement (keithley smu object).
            vg_start (float): Start voltage of transfer sweep in Volts .
            vg_stop (float): End voltage of transfer sweep in Volts.
            vg_step (float): Voltage step size for transfer sweep in Volts.
            vd_list (list): List of drain voltage steps in Volts.
            t_int (float): Integration time in sec for every data point.
            delay (float): Settling time in sec before every measurement. Set
                to -1 for for automatic delay.
            pulsed (bool): True or False for pulsed or continuous measurements.

        Returns:
            Returns a TransistorSweepData object containing sweep data.
        """
        self.busy = True
        self.abort_event.clear()

        msg = ('Recording transfer curve with Vg from %sV to %sV, Vd = %s V. '
               % (vg_start, vg_stop, vd_list))
        logger.info(msg)

        # create array with gate voltage steps, always include a step >= VgStop
        step = np.sign(vg_stop - vg_start) * abs(vg_step)
        sweeplist_gate_fwd = np.arange(vg_start, vg_stop + step, step)
        sweeplist_gate_rvs = np.flip(sweeplist_gate_fwd, 0)
        sweeplist_gate = np.append(sweeplist_gate_fwd, sweeplist_gate_rvs)

        # create ResultTable instance
        params = {'sweep_type': 'transfer', 't_int': t_int, 'delay': delay,
                  'pulsed': pulsed}
        rt = TransistorSweepData(params=params)
        rt.append_column(sweeplist_gate, name='Gate voltage', unit='V')

        # record sweeps for every drain voltage step
        for vdrain in vd_list:

            # check for abort event
            if self.abort_event.is_set():
                self.reset()
                self.beeper.beep(0.3, 2400)
                return rt

            # create array with drain voltages
            if vdrain == 'trailing':
                sweeplist_drain = sweeplist_gate
            else:
                sweeplist_drain = np.full_like(sweeplist_gate, vdrain)

            # conduct sweep
            v_g, i_g, v_d, i_d = self.voltageSweepDualSMU(
                    smu_gate, smu_drain, sweeplist_gate, sweeplist_drain, t_int,
                    delay, pulsed
                    )

            if not self.abort_event.is_set():
                i_s = np.array(i_d) + np.array(i_g)
                rt.append_column(i_s, name='Source current (Vd = %s)' % vdrain, unit='A')
                rt.append_column(i_d, name='Drain current (Vd = %s)' % vdrain, unit='A')
                rt.append_column(i_g, name='Gate current (Vd = %s)' % vdrain, unit='A')

        self.reset()
        self.beeper.beep(0.3, 2400)

        self.busy = False
        return rt

    def outputMeasurement(self, smu_gate, smu_drain, vd_start, vd_stop, vd_step,
                          vg_list, t_int, delay, pulsed):
        """
        Records an output curve and saves the results in a TransistorSweepData
        instance.

        Args:
            smu_gate: SMU attached to gate electrode of FET for transfer
                measurement (keithley smu object).
            smu_drain: SMU attached to drain electrode of FET for transfer
                measurement (keithley smu object).
            vd_start (float): Start voltage of output sweep in Volts .
            vd_stop (float): End voltage of output sweep in Volts.
            vd_step (float): Voltage step size for output sweep in Volts.
            vg_list (list): List of gate voltage steps in Volts.
            t_int (float): Integration time in sec for every data point.
            delay (float): Settling time in sec before every measurement. Set
                to -1 for for automatic delay.
            pulsed (bool): True or False for pulsed or continuous measurements.

        Returns:
            Returns a TransistorSweepData object containing sweep data.
        """

        self.busy = True
        self.abort_event.clear()
        msg = ('Recording output curve with Vd from %sV to %sV, Vg = %s V. '
               % (vd_start, vd_stop, vg_list))
        logger.info(msg)

        # create array with drain voltage steps, always include a step >= VgStop
        step = np.sign(vd_stop - vd_start) * abs(vd_step)
        sweeplist_drain_fwd = np.arange(vd_start, vd_stop + step, step)
        sweeplist_drain_rvs = np.flip(sweeplist_drain_fwd, 0)
        sweeplist_drain = np.append(sweeplist_drain_fwd, sweeplist_drain_rvs)

        # create ResultTable instance
        params = {'sweep_type': 'output', 't_int': t_int, 'delay': delay,
                  'pulsed': pulsed}
        rt = TransistorSweepData(params=params)
        rt.append_column(sweeplist_drain, name='Drain voltage', unit='V')

        for vgate in vg_list:
            if self.abort_event.is_set():
                self.reset()
                self.beeper.beep(0.3, 2400)
                return rt

            # create array with gate voltages
            sweeplist_gate = np.full_like(sweeplist_drain, vgate)

            # conduct forward sweep
            v_d, i_d, v_g, i_g = self.voltageSweepDualSMU(
                    smu_drain, smu_gate, sweeplist_drain, sweeplist_gate, t_int,
                    delay, pulsed
                    )

            if not self.abort_event.is_set():
                i_s = np.array(i_d) + np.array(i_g)
                rt.append_column(i_s, name='Source current (Vd = %s)' % vgate, unit='A')
                rt.append_column(i_d, name='Drain current (Vd = %s)' % vgate, unit='A')
                rt.append_column(i_g, name='Gate current (Vd = %s)' % vgate, unit='A')

        self.reset()
        self.beeper.beep(0.3, 2400)

        self.busy = False
        return rt

    def playChord(self, direction='up'):
        """Plays a chord on the Keithley.

        Args:
            direction (str): 'up' or 'done' for upward or downward chord
        """
        if direction is 'up':
            self.beeper.beep(0.3, 1046.5)
            self.beeper.beep(0.3, 1318.5)
            self.beeper.beep(0.3, 1568)

        elif direction is 'down':
            self.beeper.beep(0.3, 1568)
            self.beeper.beep(0.3, 1318.5)
            self.beeper.beep(0.3, 1046.5)
        else:
            self.beeper.beep(0.2, 1046.5)
            self.beeper.beep(0.1, 1046.5)


class Keithley2600Factory(object):

    _instances = {}
    SMU_LIST = Keithley2600.SMU_LIST

    def __new__(cls, *args, **kwargs):
        """
        Create new instance for a new visa_address, otherwise return existing instance.
        """
        if args[0] in cls._instances:
            logger.debug('Returning existing instance with address %s.' % args[0])

            return cls._instances[args[0]]
        else:
            logger.debug('Creating new instance with address %s.' % args[0])
            instance = Keithley2600(*args, **kwargs)
            cls._instances[args[0]] = instance

            return instance



# -*- coding: utf-8 -*-
#
# Copyright © keithley2600 Project Contributors
# Licensed under the terms of the MIT License
# (see keithley2600/__init__.py for details)

"""
Submodule defining classes to store, plot, and save measurement results.

"""

import sys
import os
import re
import time
import matplotlib.pyplot as plt
import numpy as np

PY2 = sys.version[0] == '2'

if not PY2:
    basestring = str  # in Python 3


def find_numbers(string):
    """
    Find all numbers in a string and return them in a list.
    """

    fmt = r'[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?'
    string_list = re.findall(fmt, string)
    float_list = [float(s) for s in string_list]

    return float_list


class ColumnTitle(object):
    """
    Object to hold a column title.

    :ivar name: String containing column name.
    :ivar unit: String containing column unit.
    :ivar unit_fmt: Formatting directive for column_units when generating
        string representation.
    """

    def __init__(self, name, unit=None, unit_fmt='[{}]'):
        """

        :param str name: String containing column name.
        :param str unit: String containing column unit.
        :param str unit_fmt: Formatting directive for column_units when
            generating string representation.
        """

        self.name = name
        self.unit = '' if unit is None else unit
        self.unit_fmt = unit_fmt

    def has_unit(self):
        return self.unit != ''

    def set_unit(self, unit):
        self.unit = unit

    def __repr__(self):
        return "<{0}(title='{1}', unit='{2}')>".format(
            self.__class__.__name__, self.name, self.unit)

    def __str__(self):
        if self.has_unit():
            return self.name + ' ' + self.unit_fmt.format(self.unit)
        else:
            return self.name


# noinspection PyTypeChecker
class ResultTable(object):
    """
    Class that holds measurement data. All data is stored internally in a numpy
    array.

    Columns must have column_names, to designate the measurement variable, and
    can have column_units. It is possible to access columns by their
    names in a dictionary type notation.

    :cvar COMMENT: The string used to identify a comment (default: '# ')
    :cvar DELIMITER: The string used to delimit the data (default: '\t')
    :cvar PARAM_DELIMITER: The string used to delimit (key, value) pairs (default: ': ')
    :cvar LINE_BREAK: The string used for line breaks (default: '\n' )
    :cvar UNIT_FORMAT: The format used to display column_units (default: '/{}')

    :ivar data: numpy array holding the data
    :ivar titles: list of column title objects
    :ivar params: dictionary of measurement parameters
    """

    COMMENT = '# '
    DELIMITER = '\t'
    PARAM_DELIMITER = ': '
    LINE_BREAK = '\n'
    UNIT_FORMAT = '[{}]'

    def __init__(self, names=None, units=None, data=None, params=None):
        """
        :param list names: List of column column_names (strings).
        :param list units: List of column column_units (strings).
        :param data: Numpy array holding the data.
        :param dict params: Dictionary of measurement parameters.
        """

        if names is None:
            names = []

        if units is None:
            units = [''] * len(names)

        self.titles = [ColumnTitle(n, u, self.UNIT_FORMAT) for n, u in zip(names, units)]

        if data is None:
            self.data = None
        else:
            self.data = np.array(data)

        if params is None:
            self.params = {}
        else:
            self.params = params

    @property
    def nrows(self):
        if self.data is None:
            return 0
        else:
            return self.data.shape[0]

    @property
    def ncols(self):
        return len(self.titles)

    @property
    def column_names(self):
        return [title.name for title in self.titles]

    @column_names.setter
    def column_names(self, names_list):
        if not all(isinstance(x, basestring) for x in names_list):
            raise ValueError("All column names must be of type 'str'.")
        elif not len(names_list) == self.ncols:
            raise ValueError('Number of column names does not match number of columns.')

        for title, name in zip(self.titles, names_list):
            title.name = name

    @property
    def column_units(self):
        return [title.unit for title in self.titles]

    @column_units.setter
    def column_units(self, units_list):
        if not all(isinstance(x, basestring) for x in units_list):
            raise ValueError("All column_units must be of type 'str'.")
        elif not len(units_list) == self.ncols:
            raise ValueError('Number of column_units does not match number of columns.')

        for title, unit in zip(self.titles, units_list):
            title.unit = unit

    def has_unit(self, col):
        """
        Returns `True` if column_units of column `col` have been set, `False` otherwise.

        :param col: Column index or name.

        :return: `True` if column_units have been set, `False` otherwise.
        :rtype: bool
        """
        if not isinstance(col, int):
            col = self.column_names.index(col)

        return self.titles[col].unit != ''

    def get_unit(self, col):
        """
        Get unit of column `col`.

        :param col: Column index or name (int or str)
        :return: Unit string.
        :rtype: str
        """
        if not isinstance(col, int):
            col = self.column_names.index(col)

        return self.titles[col].unit

    def set_unit(self, col, unit):
        """
        Set unit of column `col`.

        :param col: Column index or name.
        :param str unit: Unit string.
        """
        if not isinstance(col, int):
            col = self.column_names.index(col)

        self.titles[col].unit = unit

    def clear_data(self):
        """
        Clears all data.
        """
        self.data = None

    def append_row(self, data):
        """
        Appends a row to data array.

        :param data: Iterable with the same number of elements as columns in the data
            array.
        """
        if not len(data) == self.ncols:
            raise ValueError('Length must match number of columns: %s' % self.ncols)

        if self.data is None:
            self.data = np.array([data])
        else:
            self.data = np.append(self.data, [data], 0)

    def append_rows(self, data):
        """
        Appends multiple rows to data array.

        :param data: List of lists or numpy array with dimensions matching the data array.
        """

        if self.data is None:
            self.data = np.array(data)
        else:
            self.data = np.append(self.data, data, 0)

    def append_column(self, data, name, unit=None):
        """
        Appends a new column to data array.

        :param data: Iterable with the same number of elements as rows in the data array.
        :param str name: Column name.
        :param str unit: Unit of values in new column.
        """

        if self.data is None:
            self.data = np.array(np.transpose([data]))
        else:
            self.data = np.append(self.data, np.transpose([data]), 1)

        self.titles.append(ColumnTitle(name, unit, self.UNIT_FORMAT))

    def append_columns(self, data, names, units=None):
        """
        Appends multiple columns to data array.

        :param list data: List of columns to append.
        :param list names: Column column_names.
        :param list units: Units for new columns.
        """

        if self.data is None:
            self.data = np.array(np.transpose(data))
        else:
            self.data = np.append(self.data, np.transpose(data), 1)

        for name, unit in zip(names, units):
            self.titles.append(ColumnTitle(name, unit, self.UNIT_FORMAT))

    def get_row(self, i):
        """
        Returns row i as a 1D numpy array.
        """
        return self.data[i, :]

    def get_column(self, i):
        """
        Returns column i as a 1D numpy array.
        """
        return self.data[:, i]

    def column_title_string(self):
        """
        Creates column title string.

        :return: String with column titles from column column_names and column_units.
        :rtype: str
        """
        column_titles = [str(title) for title in self.titles]
        return self.DELIMITER.join(column_titles)

    def parse_column_title_string(self, title_string):
        """
        Parses a column title string and returns column column_names and column_units.

        :param str title_string: String to parse.

        :return: List of ColumnTitle objects.
        :rtype: list
        """

        title_string = title_string.lstrip(self.COMMENT)

        unique = '_UNIQUE_STRING_'
        assert unique not in self.UNIT_FORMAT

        regexp_name = '(?P<name>.*) '
        regexp_unit = re.escape(self.UNIT_FORMAT.format(unique)).replace(unique,
                                                                         '(?P<unit>.*)')

        strings = title_string.split(self.DELIMITER)

        titles = []

        for s in strings:
            m = re.search(regexp_name + regexp_unit, s)
            if m is None:
                titles.append(ColumnTitle(s, unit_fmt=self.UNIT_FORMAT))
            else:
                titles.append(ColumnTitle(m.group('name'), m.group('unit'),
                                          self.UNIT_FORMAT))

        return titles

    def param_string(self):
        """
        Creates string containing all parameters from `params` as key, value pairs in
        separate lines marked as comments.

        :return: Parameter string
        :rtype: str
        """
        lines = []

        for key, value in self.params.items():
            lines.append(str(key) + self.PARAM_DELIMITER + str(value))

        return self.LINE_BREAK.join(lines)

    def parse_param_string(self, header):
        """
        Parses comment section of header to extract measurement parameters

        :return: Dictionary containing measurement parameters.
        :rtype: dict
        """

        params = {}

        lines = header.split(self.LINE_BREAK)

        for line in lines:
            if (line.startswith(self.COMMENT) and self.PARAM_DELIMITER in line
                    and self.DELIMITER not in line):
                contents = line.lstrip(self.COMMENT)
                key, value = contents.split(self.PARAM_DELIMITER)
                try:
                    params[key] = float(value)
                except ValueError:
                    if value in ['True', 'true']:
                        params[key] = True
                    elif value in ['False', 'false']:
                        params[key] = False
                    else:
                        params[key] = value

        return params

    def header(self):
        """
        Outputs full header with comment section containing measurement parameters and
        column titles including column_units.

        :return: Header as string.
        :rtype: str
        """

        params_string = self.param_string()
        titles_string = self.column_title_string()

        return self.LINE_BREAK.join([params_string, titles_string])

    def parse_header(self, header):
        """
        Parses header. Returns list of ColumnTitle objects and measurement parameters in
        dictionary.

        :param str header: Header to parse
        :return: titles, params
        """
        header = header.strip(self.LINE_BREAK)
        last_line = header.split(self.LINE_BREAK)[-1]

        titles = self.parse_column_title_string(last_line)
        params = self.parse_param_string(header)

        return titles, params

    def save(self, filename, ext='.txt'):
        """
        Saves the result table to a text file. The file format is:
            * The header contains all measurement parameters as comments.
            * Column titles contain column_names and column_units of measured quantity.
            * Delimited columns contain the data.

        Files are saved with the specified extension (default: .txt). The
        classes default delimiters are used to separate columns and rows.

        :param str filename: Path of file to save. Relative paths are
            interpreted with respect to the current working directory.
        :param str ext: File extension (default: .txt)
        """

        base_name = os.path.splitext(filename)[0]
        filename = base_name + ext

        np.savetxt(filename, self.data, delimiter=self.DELIMITER, newline=self.LINE_BREAK,
                   header=self.header(), comments=self.COMMENT)

    def save_csv(self, filename):
        """
        Saves the result table to a csv file. The file format is:
            * The header contains all measurement parameters as comments.
            * Column titles contain column_names and column_units of measured quantity.
            * Comma delimited columns contain the data.

        Files are saved with the extension '.csv' and other extensions are
        overwritten.

        :param str filename: Path of file to save. Relative paths are
            interpreted with respect to the current working directory.
        """

        old_delim = self.DELIMITER
        old_line_break = self.LINE_BREAK
        self.DELIMITER = ','
        self.LINE_BREAK = '\n'

        self.save(filename, ext='.csv')

        self.DELIMITER = old_delim
        self.LINE_BREAK = old_line_break

    def load(self, filename):
        """
        Loads data from csv or tab delimited tex file. The header is searched
        for measurement parameters.

        :param str filename: Absolute or relative path of file to load.
        """
        old_delim = self.DELIMITER
        old_line_break = self.LINE_BREAK

        base_name, ext = os.path.splitext(filename)

        if ext == '.csv':
            self.DELIMITER = ','
            self.LINE_BREAK = '\n'

        # read info string and header
        with open(filename) as f:
            lines = f.readlines()

        header_length = sum([l.startswith(self.COMMENT) for l in lines])
        header = ''.join(lines[:header_length])

        self.titles, self.params = self.parse_header(header)

        # read data as 2D numpy array
        self.data = np.loadtxt(filename)

        self.DELIMITER = old_delim
        self.LINE_BREAK = old_line_break

    def plot(self, x_clmn=0, y_clmn=None, func=lambda x: x, **kwargs):
        """
        Plots the data. This method should not be called from a thread.
        The column containing the x-axis data is specified (defaults to first
        column), all other data is plotted on the y-axis. Keyword arguments are
        passed on to pyplot.

        Column titles are taken as legend labels. `plot` tries to determine a
        common y-axis unit and name from all given labels.

        :param x_clmn: Integer or name of column containing the x-axis data.
        :param y_clmn: List of column numbers or column_names for y-axis data. If not
            given, all columns will be plotted (excluding the x-axis column).
        :param func: Function to apply to y-data before plotting.
        """
        if self.ncols == 0:
            return

        if not isinstance(x_clmn, int):
            x_clmn = self.column_names.index(x_clmn)

        xdata = self.data[:, x_clmn]

        fig = plt.figure()

        ax = fig.add_subplot(111)

        if y_clmn is None:
            ydata = np.delete(self.data, [x_clmn], axis=1)
            ydata = func(ydata)
            lines = ax.plot(xdata, ydata, **kwargs)
            line_labels = self.column_names[0:x_clmn] + self.column_names[x_clmn + 1:]
            line_units = self.column_units[0:x_clmn] + self.column_units[x_clmn + 1:]
            ax.legend(lines, line_labels)
        else:
            line_labels = []
            line_units = []
            for c in y_clmn:
                if not isinstance(c, int):
                    c = self.column_names.index(c)
                ydata = self.data[:, c]
                ydata = func(ydata)
                ax.plot(xdata, ydata, label=self.column_names[c], **kwargs)

                line_labels.append(self.column_names[c])
                line_units.append(self.column_units[c])

        ax.set_xlabel(str(self.titles[x_clmn]))

        y_label = os.path.commonprefix(line_labels)
        y_unit = os.path.commonprefix(line_units)
        if y_unit == '':
            ax.set_ylabel('%s' % y_label)
        else:
            ax.set_ylabel(y_label + ' ' + self.UNIT_FORMAT.format(y_unit))

        ax.autoscale(enable=True, axis='x', tight=True)
        fig.tight_layout()

        self.setup_plot(fig, ax)

        fig.show()

        return fig, ax

    def setup_plot(self, fig, ax):
        """
        This method does nothing by default, but can be overridden by the child
        class in order to set up custom options for plotting.

        :param fig: Matplotlib figure instance.
        :param ax: Matplotlib axes instance.
        """
        pass

    def __repr__(self):
        titles = [str(t) for t in self.titles]
        return '<{0}(columns={1}, data=array(...))>'.format(
                self.__class__.__name__, str(titles))

# =============================================================================
# Dictionary compatibility functions. This allows access to all columns of the
# table with their names as keys.
# =============================================================================

    def keys(self):
        return self.column_names

    def has_key(self, key):
        return self.__contains__(key)

    def values(self):
        """
        Returns list of all columns.

        :return: List of columns as numpy arrays.
        :rtype: list
        """
        return [self.get_column(i) for i in range(self.ncols)]

    def items(self):
        return zip(self.keys(), self.values())

    def __getitem__(self, key):
        """
        Gets values in column with name `key`.
        :param str key: Column name.
        :return: Column content as numpy array.
        """

        if key not in self.column_names:
            raise KeyError("No such column '{0}'.".format(key))
        if not isinstance(key, str):
            raise TypeError("Key must be of type 'str'.")

        return self.get_column(self.column_names.index(key))

    def __setitem__(self, key, value):
        """
        Sets values in column with name `key`.

        :param str key: Column name.
        :param value: Iterable containing column values.
        """
        if not isinstance(key, str):
            raise TypeError("Key must be of type 'str'.")

        if key not in self.column_names:
            self.append_column(value, name=key)
        else:
            self.data[:, self.column_names.index(key)] = np.transpose(value)

    def __delitem__(self, key):
        """
        Deletes column with name `key`.
        :param str key:
        """
        i = self.column_names.index(key)
        self.data = np.delete(self.data, i, axis=1)
        self.titles.pop(i)

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n + 1 <= self.ncols:
            r = self.column_names[self._n]
            self._n += 1
            return r
        else:
            raise StopIteration

    def __contains__(self, key):
        return key in self.column_names

    def __len__(self):
        return self.ncols


class IVSweepData(ResultTable):
    """
    Class to store, load, save, and plot data from a simple IV sweep.
    """

    sweep_type = 'iv'

    def __init__(self, v=None, i=None, params=None):

        if (i is None) or (v is None):
            data = None
        else:
            data = np.transpose([v, i])

        if params is None:
            params = {}

        names = ['Voltage', 'Current']
        units = ['V', 'A']
        params['sweep_type'] = self.sweep_type
        params['recorded'] = time.localtime()

        super(self.__class__, self).__init__(names, units, data, params)

    def append(self, v, i):
        """
        Appends list-like objects with voltage and current values to data.

        :param v: List (or list-like object) containing voltages.
        :param i: List (or list-like object) containing currents.
        """

        assert len(i) == len(v)
        self.append_rows(np.transpose([v, i]))


class TransistorSweepData(ResultTable):
    """
    Class to handle, store and load transfer and output characteristic data of FETs.
    `TransistorSweepData` inherits from `ResultTable` and overrides the plot method.

    The following additional properties are accessible:
        * sweep_type: String that describes the sweep type, can be 'transfer'
          or 'output'.

    The following new methods are defined:
        * step_list: Returns list of stepped voltages.
        * n_steps: Returns number of fixed voltage steps.
    """

    @property
    def sweep_type(self):
        return self.params['sweep_type']

    @sweep_type.setter
    def sweep_type(self, value):
        self.params['sweep_type'] = value

    def stepped_voltage_list(self):
        """
        Get voltage steps of transfer / output characteristics. This returns
        the drain voltages steps for transfer curve data and gate voltage steps
        for output curve data.

        :return: Voltage steps in transfer / output characteristics.
        :rtype: set
        """

        return set(find_numbers(self.column_title_string()))

    def n_steps(self):
        return len(self.stepped_voltage_list())

    def plot(self, *args, **kwargs):

        def sqrt_abs(x):
            return np.sqrt(np.abs(x))

        if self.sweep_type == 'transfer':
            fig, ax = super(self.__class__, self).plot(func=np.abs, *args, **kwargs)
            ax.set_yscale('log')
            ax.set_ylabel('I [A]')

            fig, ax = super(self.__class__, self).plot(func=sqrt_abs, *args, **kwargs)
            ax.set_yscale('linear')
            ax.set_ylabel('$\\mathregular{I^{1/2}}$ ' +
                          self.UNIT_FORMAT.format('$\\mathregular{A^{1/2}}$'))
        elif self.sweep_type == 'output':
            fig, ax = super(self.__class__, self).plot(func=np.abs, *args, **kwargs)
            ax.set_yscale('linear')
            ax.set_ylabel('I [A]')



from setuptools import setup, find_packages

setup(name="keithley2600",
      version="1.1.0",
      description="Full Python driver for the Keithley 2600 series.",
      url="https://github.com/OE-FET/keithley2600.git",
      author="Sam Schott",
      author_email="ss2151@cam.ac.uk",
      licence='MIT',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      packages=find_packages(),
      install_requires=[
          'PyVISA',
          'pyvisa-py>=0.3.1',
          'setuptools',
          'numpy',
          'matplotlib',
      ],
      zip_safe=False,
)
